/************************************************************************/
/*                                                                      */
/*   svm_common.h                                                       */
/*                                                                      */
/*   Definitions and functions used in both svm_learn and svm_classify. */
/*                                                                      */
/*   Author: Thorsten Joachims                                          */
/*   Date: 19.07.99                                                     */
/*                                                                      */
/*   Copyright (c) 1999  Universitaet Dortmund - All rights reserved    */
/*                                                                      */
/*   This software is available for non-commercial use only. It must    */
/*   not be modified and distributed without prior permission of the    */
/*   author. The author is not responsible for implications from the    */
/*   use of this software.                                              */
/*                                                                      */
/************************************************************************/

# include <stdio.h>
# include <ctype.h>
/* # include <signal.h> */
# include <math.h>
# include <string.h>
# include <stdlib.h>
# include <sys/times.h>
/* # include <limits.h> */

# define VERSION       "V3.01"
# define VERSION_DATE  "26.10.99"

# define CFLOAT  float       /* the type of float to use for caching */
                             /* kernel evaluations. Using float saves */
                             /* us some memory, but you can use double, too */
# define FNUM    long        /* the type used for storing feature ids */
# define FVAL    float       /* the type used for storing feature values */

# define LINEAR  0           /* linear kernel type */
# define POLY    0           /* polynoial kernel type */
# define RBF     0           /* rbf kernel type */
# define SIGMOID 0           /* sigmoid kernel type */

typedef struct word {
  FNUM    wnum;	
  FVAL    weight;
} WORD;

typedef struct doc {
  long    docnum;
  double  twonorm_sq;
  WORD    *words;
} DOC;

typedef struct learn_parm {
  long   svm_c_steps;          /* do so many steps for finding optimal C */
  double svm_c_factor;         /* increase C by this factor every step */
  double svm_c;
  double svm_costratio;
  double svm_costratio_unlab;
  double svm_unlabbound;
  double *svm_cost;            /* individual upper bounds for each var */
  double transduction_posratio;/* fraction of unlabeled examples to be */
                               /* classified as positives */
  long   remove_inconsistent;
  long   biased_hyperplane;
  long   skip_final_opt_check;
  long   svm_maxqpsize;
  long   svm_iter_to_shrink;
  double epsilon_crit;         /* tolerable error for distances used */
                               /* in stopping criterion */
  double epsilon_a;            /* tolerable error on alphas at bounds */
  double epsilon_shrink;       /* how much a multiplier should be above */
                               /* zero for shrinking */
  double epsilon_const;        /* tolerable error on eq-constraint */
  double opt_precision;        /* precision of solver, set to 1e-21 */
                               /* if you get convergence problems */
  char predfile[200];          /* file for classes of unlabeled examples */
                               /* in case of transduction */
  char alphafile[200];         /* file to store optimal alphas in */
} LEARN_PARM;

typedef struct kernel_parm {
  long    kernel_type;   /* 0=linear, 1=poly, 2=rbf, 3=sigmoid, 4=custom */
  long    poly_degree;
  double  rbf_gamma;
  double  coef_lin;
  double  coef_const;
  char    custom[50];    /* for user supplied kernel */
} KERNEL_PARM;

typedef struct model {
  long    sv_num;	
  long    at_upper_bound;
  double  b;
  DOC     **supvec;
  double  *alpha;
  long    *index;       /* index from docnum to position in model */
  double  *lin_weights; /* weights for linear case using folding */
} MODEL;

typedef struct quadratic_program {
  long   opt_n;            /* number of variables */
  long   opt_m;            /* number of linear equality constraints */
  double *opt_ce,*opt_ce0; /* linear equality constraints */
  double *opt_g;           /* hessian of objective */
  double *opt_g0;          /* linear part of objective */
  double *opt_xinit;       /* initial value for variables */
  double *opt_low,*opt_up; /* box constraints */
} QP;

typedef struct kernel_cache {
  long   *index;  /* cache some kernel evalutations */
  CFLOAT *buffer; /* to improve speed */
  long   *invindex;
  long   *active2totdoc;
  long   *totdoc2active;
  long   *lru;
  long   *occu;
  long   elems;
  long   max_elems;
  long   time;
  long   activenum;
  long   buffsize;
} KERNEL_CACHE;


double classify_example(MODEL *, DOC *, KERNEL_PARM *);
CFLOAT kernel(KERNEL_PARM *, DOC *, DOC *); 
double custom_kernel(KERNEL_PARM *, DOC *, DOC *); 
double sprod_ss(WORD *, WORD *);
double model_length_s(MODEL *, KERNEL_PARM *);
void   clear_vector_n(double *, long);
void   add_vector_ns(double *, WORD *, double);
double sprod_ns(double *, WORD *);
void   read_documents(char *, DOC *, long *, long, long, long *, long *);
int    parse_document(char *, DOC *, long *, long *, long);
void   nol_ll(char *, long *, long *, long *);
long   min(long, long);
long   max(long, long);
long   get_runtime();
void   *my_malloc(long); 
void   copyright_notice();

long   verbosity;              /* verbosity level (0-4) */
long   kernel_cache_statistic;


